import requests
import re
from user_agents import parse
from functools import reduce
from django.core.exceptions import ImproperlyConfigured
from django.utils.deprecation import MiddlewareMixin
from django.conf import settings


class TrackerMiddleware(MiddlewareMixin):
    def __init__(self, get_response):
        if not hasattr(settings, "ACKEE_SERVER"):
            raise ImproperlyConfigured("Please set ACKEE_SERVER in settings")
        if not hasattr(settings, "ACKEE_DOMAIN_ID"):
            raise ImproperlyConfigured("Please set ACKEE_DOMAIN_ID in settings")
        if not hasattr(settings, "ACKEE_IGNORED_PATHS"):
            raise ImproperlyConfigured("Please set ACKEE_IGNORED_PATHS in settings")
        self.get_response = get_response

    def _send(self, data):
        url = f"{settings.ACKEE_SERVER}/domains/{settings.ACKEE_DOMAIN_ID}/records"
        response = requests.post(url, json=data)
        if response.status_code == 202:
            return response.json()
        else:
            raise Exception(response.text)

    def _parse_accept_language(self, accept_language=""):
        """Parse and sort the Accept-Language header
        Taken from https://siongui.github.io/2012/10/11/python-parse-accept-language-in-http-request-header/

        Arguments:
            accept_language {str} -- Accept-Language header

        Returns:
            list -- A list of touples containing pairs of language and its priority
        """
        if accept_language is None:
            return []

        languages = accept_language.split(",")
        locale_q_pairs = []

        if len(languages) == 1 and languages[0] == "":
            return []

        for language in languages:
            if language.split(";")[0] == language:
                # no q => q = 1
                locale_q_pairs.append((language.strip(), "1"))
            else:
                locale = language.split(";")[0].strip()
                q = language.split(";")[1].split("=")[1]
                locale_q_pairs.append((locale, q))

        return locale_q_pairs

    def _sanitize_accept_language(self, accept_language):
        languages = self._parse_accept_language(accept_language)
        if len(languages) == 0:
            return None
        return languages[0][0][:2]

    def _is_ignored_path(self, path):
        for ignore in settings.ACKEE_IGNORED_PATHS:
            if re.match(ignore, path) is not None:
                return True
        return False

    def process_request(self, request):
        # Do not track users who specified DNT header
        if int(request.headers.get("DNT", "0")) == 1:
            return

        if self._is_ignored_path(request.get_full_path()):
            return

        user_agent = parse(request.headers.get("User-Agent", ""))

        data = {
            "siteLocation": request.build_absolute_uri(),
            "siteReferrer": request.headers.get("Referer"),
            "siteLanguage": self._sanitize_accept_language(
                request.headers.get("Accept-Language")
            ),
            "deviceName": user_agent.device.model,
            "deviceManufacturer": user_agent.device.brand,
            "osName": user_agent.os.family,
            "osVersion": user_agent.os.version_string,
            "browserName": user_agent.browser.family,
            "browserVersion": user_agent.browser.version_string,
        }

        try:
            self._send(data)
        except:
            pass



import pytest
from unittest.mock import patch, MagicMock
from django.core.exceptions import ImproperlyConfigured
from ackee.middleware import TrackerMiddleware
from django.test.utils import override_settings

default_settings = {
    "ACKEE_SERVER": "foo",
    "ACKEE_DOMAIN_ID": 123,
    "ACKEE_IGNORED_PATHS": [],
}


class TestInitialization:
    @patch("ackee.middleware.settings")
    def test_throws_when_no_server(self, settings):
        del settings.ACKEE_SERVER
        with pytest.raises(ImproperlyConfigured):
            TrackerMiddleware("foo")

    @patch("ackee.middleware.settings")
    def test_throws_when_no_domain_id(self, settings):
        del settings.ACKEE_DOMAIN_ID
        with pytest.raises(ImproperlyConfigured):
            TrackerMiddleware("foo")

    @patch("ackee.middleware.settings")
    def test_throws_when_no_ignored_paths(self, settings):
        del settings.ACKEE_IGNORED_PATHS
        with pytest.raises(ImproperlyConfigured):
            TrackerMiddleware("foo")


class TestMiddleware:
    class TestSending:
        @override_settings(**default_settings)
        @patch("ackee.middleware.requests.post")
        def test_returns_response(self, post):
            tracker = TrackerMiddleware("foo")
            response = MagicMock()
            response.status_code = 202
            response.json.side_effect = ["foo"]
            post.side_effect = [response]

            result = tracker._send({"bar": "baz"})
            post.assert_called_once_with("foo/domains/123/records", json={"bar": "baz"})
            assert result == "foo"

        @override_settings(**default_settings)
        @patch("ackee.middleware.requests.post")
        def test_throws_on_error(self, post):
            tracker = TrackerMiddleware("foo")
            response = MagicMock()
            response.status_code = 500
            response.text = "bar"
            post.side_effect = [response]

            with pytest.raises(Exception):
                result = tracker._send({})

    class TestParsing:
        @override_settings(**default_settings)
        def test_returns_list_of_touples(self):
            tracker = TrackerMiddleware("foo")
            result = tracker._parse_accept_language("en-GB,en-US;q=0.9,en;q=0.8")
            assert result == [
                ("en-GB", "1"),
                ("en-US", "0.9"),
                ("en", "0.8"),
            ]
            assert tracker._parse_accept_language() == []
            assert tracker._parse_accept_language(None) == []

    class TestSanitizing:
        @override_settings(**default_settings)
        def test_returns_none(self):
            tracker = TrackerMiddleware("foo")
            assert tracker._sanitize_accept_language("") == None

        @override_settings(**default_settings)
        def test_returns_first_language(self):
            tracker = TrackerMiddleware("foo")
            assert (
                tracker._sanitize_accept_language("en-GB,en-US;q=0.9,en;q=0.8") == "en"
            )

    class TestIgnoringPaths:
        @override_settings(**default_settings)
        def test_returns_bool(self):
            tracker = TrackerMiddleware("foo")
            assert tracker._is_ignored_path("/foo") == False
            with patch("ackee.middleware.settings.ACKEE_IGNORED_PATHS", "^/bar"):
                assert tracker._is_ignored_path("/bar") == True

    class TestProcessingRequest:
        def setup_method(self, method):
            with override_settings(**default_settings):
                self.tracker = TrackerMiddleware("foo")
                self.tracker._send = MagicMock()
                self.tracker._is_ignored_path = MagicMock()
                self.tracker._is_ignored_path.return_value = False

                self.request = MagicMock()
                self.request.headers = {}

        def test_skips_when_dnt(self):
            self.request.headers = {"DNT": "1"}
            self.tracker.process_request(self.request)
            self.tracker._send.assert_not_called()

        def test_skips_when_ignored(self):
            self.tracker._is_ignored_path.return_value = True
            self.request.get_full_path.return_value = "/foo"

            self.tracker.process_request(self.request)

            self.tracker._is_ignored_path.assert_called_once_with("/foo")
            self.tracker._send.assert_not_called()

        def test_sends_data(self):
            self.request.headers = {
                "User-Agent": "Mozilla/5.0 (iPhone; CPU iPhone OS 13_3_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.5 Mobile/15E148 Safari/604.1",
                "Referer": "http://example.com",
                "Accept-Language": "en-GB,en-US;q=0.9,en;q=0.8",
            }
            self.request.build_absolute_uri.return_value = "http://foo.bar"
            self.tracker._send.side_effect = Exception()

            self.tracker.process_request(self.request)

            self.tracker._send.assert_called_once_with(
                {
                    "siteLocation": "http://foo.bar",
                    "siteReferrer": "http://example.com",
                    "siteLanguage": "en",
                    "deviceName": "iPhone",
                    "deviceManufacturer": "Apple",
                    "osName": "iOS",
                    "osVersion": "13.3.1",
                    "browserName": "Mobile Safari",
                    "browserVersion": "13.0.5",
                }
            )
